/*!\file Dofx
 * \brief: establish degrees of freedom for all nodes
 */

#include "./Dofx.h"

#include "../shared/shared.h"
#include "../include/macros.h"
#include "../toolkits/toolkits.h"
#include "../EnumDefinitions/EnumDefinitions.h"

int Dofx( DofVec** ppartition, DofVec** ptpartition,DataSet* elements,DataSet* nodes, DataSet* params) {

	int noerr=1;
	int i;

	int  found=0;
	extern int num_procs;
	extern int my_rank;
	
	/*intermediary: */
	int  numberofnodes;
	int  numberofdofspernode;

	/*output: */
	DofVec* partition=NULL;
	DofVec* tpartition=NULL;

	/*Initialize dofvecs: */
	partition=new DofVec("partition");
	tpartition=new DofVec("tpartition");

	/*First, recover number of grids from parameters: */
	found=params->FindParam(&numberofnodes,"numberofnodes");
	if(!found)ISSMERROR(" could not find numberofnodes in parameters");

	/*Recover number of dofs per node: */
	found=params->FindParam(&numberofdofspernode,"numberofdofspernode");
	if(!found)ISSMERROR(" could not find numberofdofspernode in parameters");

	/*Ensure that only for each cpu, the partition border nodes only will be taken into account once 
	 * across the cluster. To do so, we flag all the clone nodes: */
	nodes->FlagClones(numberofnodes);

	/*Go through all nodes, and build degree of freedom lists. Each node gets a fixed number of dofs. When 
	 *a  node has already been distributed dofs on one cpu, all other cpus with the same node cannot distribute it 
	 *anymore. Use clone field to be sure of that: */
	nodes->DistributeDofs(numberofnodes,numberofdofspernode);

	/*Now that dofs have been distributed, create partitioning vector and its transpose: */
	nodes->CreatePartitioningVector(&partition->vector,numberofnodes,numberofdofspernode);

	/*Transpose partition into tpartition: */
	VecTranspose(&tpartition->vector,partition->vector);

	/*Now, setup numdof and numentries for dofvec: */
	partition->numdofs=1;
	VecGetSize(partition->vector,&partition->numentries);

	tpartition->numdofs=1;
	VecGetSize(tpartition->vector,&tpartition->numentries);

	/*Assign output pointers: */
	*ppartition=partition;
	*ptpartition=tpartition;
	
	return noerr;
}
