/*! \file CreateLoadsMelting.c:
 */

#include "../../DataSet/DataSet.h"
#include "../../toolkits/toolkits.h"
#include "../../EnumDefinitions/EnumDefinitions.h"
#include "../../objects/objects.h"
#include "../../shared/shared.h"
#include "../../include/macros.h"
#include "../IoModel.h"


void	CreateLoadsMelting(DataSet** ploads, IoModel* iomodel,ConstDataHandle iomodel_handle){

	int i,j,counter;
	int element;

	extern int my_rank;
	extern int num_procs;
	
	DataSet*    loads    = NULL;
	Pengrid*    pengrid  = NULL;

	int segment_width;
	int i1,i2,i3,i4;
	int i0,range;
	int grid1,grid2;

	/*pengrid intermediary data: */
	int pengrid_id;
	int pengrid_node_id;
	int pengrid_mparid;
	int pengrid_dof;
	double pengrid_penalty_offset;
	int pengrid_active=0;
	int pengrid_thermal_steadystate=1;
	int pengrid_stabilize_constraints=0;

	int count;

	/*Create loads: */
	loads   = new DataSet(LoadsEnum());
	count=0;

	/*if 2d: return*/
	if (strcmp(iomodel->meshtype,"2d")==0)goto cleanup_and_return;

	//create penalties for grids: no grid can have a temperature over the melting point
	IoModelFetchData(&iomodel->gridonbed,NULL,NULL,iomodel_handle,"gridonbed");

	for (i=0;i<iomodel->numberofnodes;i++){
	#ifdef _PARALLEL_
	/*keep only this partition's nodes:*/
	if((iomodel->my_grids[i]==1)){
	#endif

		if (iomodel->gridonbed[i]){ 
		
			pengrid_id=count+1; //matlab indexing
			pengrid_node_id=i+1;
			pengrid_dof=1;
			pengrid_penalty_offset=iomodel->penalty_offset;
			pengrid_active=0;
			pengrid_mparid=iomodel->numberofelements+1;//refers to the corresponding parmat property card
			pengrid_stabilize_constraints=iomodel->stabilize_constraints;
			
			pengrid= new Pengrid(pengrid_id, pengrid_node_id,pengrid_mparid,pengrid_dof, pengrid_active, pengrid_penalty_offset,pengrid_thermal_steadystate,pengrid_stabilize_constraints);
			
			loads->AddObject(pengrid);
			count++;
		}
	#ifdef _PARALLEL_
	} //if((iomodel->my_grids[i]==1))
	#endif
	}
	xfree((void**)&iomodel->gridonbed);

	/*All our datasets are already order by ids. Set presort flag so that later on, when sorting is requested on these 
	 * datasets, it will not be redone: */
	loads->Presort();

	cleanup_and_return:

	/*Assign output pointer: */
	*ploads=loads;

}


