/*!\file UpdateNodePositionsx
 * \brief update node positions using new geometry defined by new thickness and new bed.
 */

#include "./UpdateNodePositionsx.h"

#include "../shared/shared.h"
#include "../include/macros.h"
#include "../toolkits/toolkits.h"
#include "../EnumDefinitions/EnumDefinitions.h"

void UpdateNodePositionsx( DataSet* elements,DataSet* nodes,DataSet* loads, DataSet* materials, DataSet* parameters,Vec thickness,Vec bed){

	/*intermediary: */
	double* thickness_serial=NULL;
	double* bed_serial=NULL;

	/*First, get elements and loads configured: */
	elements->Configure(elements,loads, nodes, materials,parameters);
	loads->Configure(elements, loads, nodes, materials,parameters);
	nodes->Configure(elements, loads, nodes, materials,parameters);
	parameters->Configure(elements,loads, nodes, materials,parameters);

	/*serialize inputs: */
	VecToMPISerial(&thickness_serial,thickness);
	VecToMPISerial(&bed_serial,bed);
	
	/*Update elements, nodes, loads and materials from new geometry: */
	nodes->UpdateNodePositions(thickness_serial,bed_serial);

	/*Free ressources:*/
	xfree((void**)&thickness_serial);
	xfree((void**)&bed_serial);
}
